#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""The SMILES of two small molecules for the receptors with PDB id: : 6lu7, x0161.
"""
r_6lu7 = "CC(C)C[C@H](NC(=O)[C@@H](NC(=O)[C@H](C)NC(O)C1CC(C)ON1)C(C)C)C(=O)N[C@H](CC[C@H](O)OCC1CCCCC1)C[C@@H]1CCNC1O"
r_x0161 = "COC(=O)C=1C=CC(=CC1)S(=O)(=O)N"

