#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# This is a dummy file. It is reserved just for Fitness function implemented for the user and the command line.
# The only function is to avoid warnings in the code when CustomMolDrugFitness is imported in cli.py